#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author : Murphy
# LICENSE: GPL v3

import os
import shutil
import zipfile
from r2flutch.repl import print_console


def generate_ipa(path, app_name, target_dir='.'):
    ipa_filename = app_name + '.ipa'
    output_path = os.path.join(target_dir, ipa_filename)
    print_console('Creating IPA file at %s' % output_path)
    with zipfile.ZipFile(output_path, mode='w', compression=zipfile.ZIP_DEFLATED) as zip_handle:
        for (root, _, files) in os.walk(path):
            for file in files:
                zip_handle.write(os.path.join(root, file),
                                 os.path.relpath(os.path.join(root, file), os.path.join(path, '..')))
    print_console('IPA file saved at %s' % output_path)


def copy_modules_to_app_bundle(dumped_modules, app_path):
    for module in dumped_modules:
        if os.path.dirname(module['relative_path']):
            os.makedirs(os.path.dirname(module['relative_path']), exist_ok=True)
        shutil.move(module['src_path'], os.path.join(app_path, module['relative_path']))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author : Murphy
# LICENSE: GPL v3

import argparse
import os
import shutil
import sys
from r2flutch.repl import print_console, SUCCESS
from r2flutch.device import get_usb_device, list_applications, spawn_r2frida_process
from r2flutch.io import set_block_size, mount_app_bundle, copy_application_bundle, list_application_content
from r2flutch.r2frida import load_all_modules, dump_decrypted_modules, get_main_bundle_name, load_r2f_plugin
from r2flutch.config import BLOCKSIZE, TMP_FOLDER, DUMP_FOLDER, BIN_FOLDER
from r2flutch.utils import generate_ipa, copy_modules_to_app_bundle


def init():
    if os.path.isdir(TMP_FOLDER):
        shutil.rmtree(TMP_FOLDER)
    os.makedirs(TMP_FOLDER)
    os.makedirs(os.path.join(TMP_FOLDER, DUMP_FOLDER))
    os.makedirs(os.path.join(TMP_FOLDER, BIN_FOLDER))


def main():
    parser = argparse.ArgumentParser(description='r2flutch (by Murphy)')
    parser.add_argument('-d', '--debug', dest='debug', action='store_true', help='Show debug messages')
    parser.add_argument('-o', '--output', type=str, help='Path where output files will be stored.')
    parser.add_argument('-i', '--ipa', dest='generate_ipa', action='store_true', help='Generate an IPA file')
    parser.add_argument('-l', '--list', dest='list', action='store_true', help='List the installed apps')
    parser.add_argument('target', nargs='?', help='Bundle identifier of the target app')
    arguments = parser.parse_args()
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        sys.exit(0)
    init()
    output_dir = arguments.output if arguments.output else '.'
    os.makedirs(output_dir, exist_ok=True)
    debug_enabled = arguments.debug
    device = get_usb_device()
    if arguments.list:
        list_applications(device)
        sys.exit(0)
    r2f = spawn_r2frida_process(arguments.target, device.id)
    mount_app_bundle(r2f)
    load_r2f_plugin(r2f)
    set_block_size(r2f, BLOCKSIZE)
    app_content = list_application_content(r2f, '')
    load_all_modules(r2f, app_content)
    dumped_modules = dump_decrypted_modules(r2f, debug_enabled)
    app_name = get_main_bundle_name(r2f)
    if arguments.generate_ipa:
        payload_path = os.path.join(TMP_FOLDER, 'Payload')
        app_path = os.path.join(payload_path, '%s.app' % app_name)
        os.makedirs(app_path, exist_ok=True)
        copy_application_bundle(r2f, app_content, app_path, debug_enabled)
        copy_modules_to_app_bundle(dumped_modules, app_path)
        generate_ipa(payload_path, app_name, target_dir=output_dir)
    else:
        for module in dumped_modules:
            shutil.move(module['src_path'], output_dir)
        print_console('Decrypted modules saved at %s' % output_dir)
    if dumped_modules:
        end_msg = 'r2flutch Decryption Complete!'
        print_console(end_msg, level=SUCCESS)
        r2f.cmd(':?E ' + end_msg)
    r2f.quit()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author : Murphy
# LICENSE: GPL v3

import sys
import threading
import frida
import r2pipe
from r2flutch.repl import print_console, ERROR, DEFAULT


def get_usb_device():
    device_type = 'usb'
    device_manager = frida.get_device_manager()
    changed = threading.Event()

    def on_changed():
        changed.set()

    device_manager.on('changed', on_changed)
    device = None
    while device is None:
        devices = [dev for dev in device_manager.enumerate_devices() if dev.type == device_type]
        if len(devices) == 0:
            print_console('Waiting for USB device...')
            changed.wait()
        else:
            device = devices[0]
    device_manager.off('changed', on_changed)
    return device


def list_applications(device):
    try:
        applications = device.enumerate_applications()
    except Exception as err:
        sys.exit('[x] Failed to enumerate applications: %s' % err)
    print_console('Bundle Identifier', level=DEFAULT)
    print_console('------------------------------', level=DEFAULT)
    for application in applications:
        print_console(application.identifier, level=DEFAULT)


def spawn_r2frida_process(bundle_id, device_id):
    print_console('Open Application Process %s' % bundle_id)
    r2frida_handle = 'frida://launch/usb/%s/%s' % (device_id, bundle_id)
    try:
        r2f = r2pipe.open(r2frida_handle)
        return r2f
    except Exception:
        print_console('Cannot open target process: ' + bundle_id, ERROR)
        sys.exit(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author : Murphy
# LICENSE: GPL v3

import os
import shutil
import base64
from tqdm import tqdm
from r2flutch.repl import print_console, ERROR


PREFIX = '/r2f'
APP_BUNDLE_BASE_PATH = '%s/AppBundle' % PREFIX
DEVICE_PATH = '%s/Device' % PREFIX


def set_block_size(r2f, blocksize):
    r2f.cmd(r'b %s' % blocksize)
    print_console('Set block size to ' + r2f.cmd(r'b'))


def mount_app_bundle(r2f):
    print_console('Mount Application Bundle')
    r2f.cmd(r'm %s io 0x0' % PREFIX)


def copy_application_bundle(r2f, app_content, dest, debug_enabled):
    print_console('Copy application bundle to: %s' % dest)
    if not os.path.exists(dest):
        os.makedirs(dest)
    print_console('Copy App Bundle to disk')
    for filepath in tqdm(app_content):
        get_file(r2f, filepath, dest, debug_enabled)


def download_module(r2f, module_path, dest_path):
    basename = os.path.basename(module_path)
    full_path = os.path.join(DEVICE_PATH, module_path[1:])
    r2f.cmd('\"mg %s\"' % full_path)
    shutil.move(basename, dest_path)


def get_file(r2f, filepath, dest, debug_enabled):
    dirs = os.path.dirname(filepath)
    basename = os.path.basename(filepath)
    if dirs:
        os.makedirs(os.path.join(dest, dirs), exist_ok=True)
    b64_full_path = base64.b64encode(os.path.join(APP_BUNDLE_BASE_PATH, filepath).encode()).decode()
    if debug_enabled:
        r2f.cmd('mg base64:%s' % b64_full_path)
    else:
        r2f.cmd('mg base64:%s 2>/dev/null' % b64_full_path)
    if os.path.isfile(basename):
        shutil.move(basename, os.path.join(dest, dirs, basename))
    else:
        print_console('Failed to copy file: %s' % basename, level=ERROR)


def list_content_path(r2f, path):
    b64_full_path = base64.b64encode(os.path.join(APP_BUNDLE_BASE_PATH, path).encode()).decode()
    r2cmd = 'mdj base64:%s' % b64_full_path
    return r2f.cmdj(r2cmd)


def list_application_content(r2f, path):
    app_content = []
    for file_obj in list_content_path(r2f, path):
        if file_obj['type'] == 'file':
            file_path = os.path.join(path, file_obj['name'])
            app_content.append(file_path)
        else:
            app_content.extend(list_application_content(r2f, os.path.join(path, file_obj['name'])))
    return app_content



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author : Murphy
# LICENSE: GPL v3

import os
import re
import r2pipe
import r2flutch
from r2flutch.repl import print_console, DEBUG
from r2flutch.io import APP_BUNDLE_BASE_PATH, download_module
from r2flutch.config import TMP_FOLDER, DUMP_FOLDER, BIN_FOLDER


def load_all_modules(r2f, app_content):
    print_console('Loading all modules')
    for filepath in app_content:
        ext = os.path.splitext(filepath)[1]
        if ext == '.dylib':
            load_library(r2f, os.path.join(APP_BUNDLE_BASE_PATH, filepath))
        elif ext == '.framework':
            load_library(r2f, os.path.join(APP_BUNDLE_BASE_PATH, filepath))


def load_library(r2f, path):
    print_console('Loading library %s' % path)
    r2cmd = ':dl %s' % path
    return r2f.cmdj(r2cmd)


def load_framework(r2f, path):
    print_console('Loading framework %s' % path)
    r2cmd = ':dlf %s' % path
    return r2f.cmdj(r2cmd)


def load_r2f_plugin(r2f):
    plugin_path = os.path.join(os.path.dirname(os.path.abspath(r2flutch.__file__)), 'agent', 'plugin.js')
    r2f.cmd(':. %s' % plugin_path)


def get_main_bundle_name(r2f):
    return r2f.cmd(':getMainBundleName').strip()


def dump_decrypted_modules(r2f, debug_enabled=False):
    dumped_modules = []
    modules_to_decrypt = r2f.cmdj(':dump')
    for module in modules_to_decrypt:
        bin_dumped = os.path.join(TMP_FOLDER, DUMP_FOLDER, module['name'])
        dst_bin = os.path.join(TMP_FOLDER, BIN_FOLDER, module['name'])
        crypt_offset = hex(module['encryption_info']['cryptoff'])
        base_addr = hex(int(module['base'], 16))
        offset = hex(int(module['base'], 16) + module['encryption_info']['cryptoff'])
        size = hex(module['encryption_info']['cryptsize'])
        crypt_header = hex(int(module['encryption_info']['addr'], 16))
        print_console('Dumping Module %s at %s (%s Bytes)' % (module['name'], offset, size))
        if debug_enabled:
            print_console('BaseAddr: %s' % base_addr, level=DEBUG)
            print_console('CryptOff: %s' % crypt_offset, level=DEBUG)
            print_console('BaseAddr + CryptOff: %s' % offset, level=DEBUG)
            print_console('CryptSize: %s' % size, level=DEBUG)
            print_console('Crypt Header: %s' % crypt_header, level=DEBUG)
        r2f.cmd('s %s' % offset)
        r2f.cmd('wtf %s %s' % (bin_dumped, size))
        download_module(r2f, module['path'], dst_bin)
        patch_bin(bin_dumped, dst_bin, crypt_header, crypt_offset, debug_enabled)
        print_console('Module %s successfully decrypted' % dst_bin)
        relative_path = re.sub(r".*\.app/", "", module['path'])
        dumped_modules.append({'src_path': dst_bin, 'relative_path': relative_path})
    return dumped_modules


def patch_bin(bin_dumped, dst_bin, crypt_header, cryptoff, debug_enabled=False):
    r2_raw = r2pipe.open(dst_bin, ['-nw'])
    cryptid_offset = hex(int(crypt_header, 16) + int('0x10', 16))
    if debug_enabled:
        print_console('dst_bin: %s' % dst_bin, level=DEBUG)
        print_console('crypt_header: %s' % crypt_header, level=DEBUG)
        print_console('crypt_offset: %s' % cryptoff, level=DEBUG)
        print_console('cryptid_offset: %s' % cryptid_offset, level=DEBUG)
    print_console('Writing decrypted data to file %s at %s' % (dst_bin, cryptoff))
    r2_raw.cmd('s %s; wff %s' % (cryptoff, bin_dumped))
    print_console('Patching cryptid at offset %s' % cryptid_offset)
    r2_raw.cmd('s %s; wx 00' % cryptid_offset)
    r2_raw.quit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author : Murphy
# LICENSE: GPL v3

from colorama import Fore, Style

# Global Vars
INFO = 0
SUCCESS = 1
ERROR = 2
DEBUG = 3
DEFAULT = 4


def print_console(msg, level=INFO, formatter=0):
    tabs, color = ["", ""]
    for _ in range(formatter):
        tabs += "    "
    if level == ERROR:
        msg = "[x] ERROR - " + msg
        color = Fore.RED
    elif level == SUCCESS:
        msg = "[*] SUCCESS - " + msg
        color = Fore.YELLOW
    elif level == DEBUG:
        msg = "[DEBUG] - " + msg
        color = Fore.BLUE
    elif level == INFO:
        msg = "[+] " + msg
    print(color + tabs + msg + Style.RESET_ALL)
